package at.ac.uibk.gitsearch.service;

// for static metamodels
import at.ac.uibk.gitsearch.domain.ReviewHistoryEntry;
import at.ac.uibk.gitsearch.domain.ReviewHistoryEntry_;
import at.ac.uibk.gitsearch.domain.ReviewRating_;
import at.ac.uibk.gitsearch.domain.Review_;
import at.ac.uibk.gitsearch.domain.User_;
import at.ac.uibk.gitsearch.repository.jpa.ReviewHistoryEntryJPARepository;
import at.ac.uibk.gitsearch.service.criteria.ReviewHistoryEntryCriteria;
import at.ac.uibk.gitsearch.service.dto.ReviewHistoryEntryDTO;
import at.ac.uibk.gitsearch.service.mapper.ReviewHistoryEntryMapper;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import javax.persistence.criteria.JoinType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import tech.jhipster.service.QueryService;

/**
 * Service for executing complex queries for {@link ReviewHistoryEntry} entities in the database.
 * The main input is a {@link ReviewHistoryEntryCriteria} which gets converted to {@link Specification},
 * in a way that all the filters must apply.
 * It returns a {@link List} of {@link ReviewHistoryEntryDTO} or a {@link Page} of {@link ReviewHistoryEntryDTO} which fulfills the criteria.
 */
@Service
@Transactional(readOnly = true)
@SuppressWarnings({ "PMD.CyclomaticComplexity", "PMD.NPathComplexity" })
public class ReviewHistoryEntryQueryService extends QueryService<ReviewHistoryEntry> {

    private final Logger log = LoggerFactory.getLogger(ReviewHistoryEntryQueryService.class);
    private final ReviewHistoryEntryJPARepository reviewHistoryEntryRepository;

    private final ReviewHistoryEntryMapper reviewHistoryEntryMapper;

    public ReviewHistoryEntryQueryService(
        ReviewHistoryEntryJPARepository reviewHistoryEntryRepository,
        ReviewHistoryEntryMapper reviewHistoryEntryMapper
    ) {
        super();
        this.reviewHistoryEntryRepository = reviewHistoryEntryRepository;
        this.reviewHistoryEntryMapper = reviewHistoryEntryMapper;
    }

    /**
     * Return a {@link List} of {@link ReviewHistoryEntryDTO} which matches the criteria from the database.
     * @param criteria The object which holds all the filters, which the entities should match.
     * @return the matching entities.
     */
    @Transactional(readOnly = true)
    public List<ReviewHistoryEntryDTO> findByCriteria(ReviewHistoryEntryCriteria criteria) {
        log.debug("find by criteria : {}", criteria);
        final Specification<ReviewHistoryEntry> specification = createSpecification(criteria);
        return reviewHistoryEntryMapper.toDto(reviewHistoryEntryRepository.findAll(specification, createSort(criteria)));
    }

    /**
     * Return a {@link Page} of {@link ReviewHistoryEntryDTO} which matches the criteria from the database.
     * @param criteria The object which holds all the filters, which the entities should match.
     * @param page The page, which should be returned.
     * @return the matching entities.
     */
    @Transactional(readOnly = true)
    public Page<ReviewHistoryEntryDTO> findByCriteria(ReviewHistoryEntryCriteria criteria, Pageable page) {
        log.debug("find by criteria : {}, page: {}", criteria, page);
        final Specification<ReviewHistoryEntry> specification = createSpecification(criteria);
        return reviewHistoryEntryRepository.findAll(specification, page).map(reviewHistoryEntryMapper::toDto);
    }

    /**
     * Return the number of matching entities in the database.
     * @param criteria The object which holds all the filters, which the entities should match.
     * @return the number of matching entities.
     */
    @Transactional(readOnly = true)
    public long countByCriteria(ReviewHistoryEntryCriteria criteria) {
        log.debug("count by criteria : {}", criteria);
        final Specification<ReviewHistoryEntry> specification = createSpecification(criteria);
        return reviewHistoryEntryRepository.count(specification);
    }

    /**
     * Function to convert {@link ReviewHistoryEntryCriteria} to a {@link Specification}
     * @param criteria The object which holds all the filters, which the entities should match.
     * @return the matching {@link Specification} of the entity.
     */
    @SuppressWarnings("PMD.CognitiveComplexity")
    protected Specification<ReviewHistoryEntry> createSpecification(ReviewHistoryEntryCriteria criteria) {
        Specification<ReviewHistoryEntry> specification = Specification.where(null);
        if (criteria != null) {
            // This has to be called first, because the distinct method returns null
            if (criteria.getDistinct() != null) {
                specification = specification.and(distinct(criteria.getDistinct()));
            }
            if (criteria.getId() != null) {
                specification = specification.and(buildRangeSpecification(criteria.getId(), ReviewHistoryEntry_.id));
            }
            if (criteria.getMessage() != null) {
                specification = specification.and(buildStringSpecification(criteria.getMessage(), ReviewHistoryEntry_.message));
            }
            if (criteria.getType() != null) {
                specification = specification.and(buildSpecification(criteria.getType(), ReviewHistoryEntry_.type));
            }
            if (criteria.getEntryDate() != null) {
                specification = specification.and(buildRangeSpecification(criteria.getEntryDate(), ReviewHistoryEntry_.entryDate));
            }
            if (criteria.getReviewId() != null) {
                specification =
                    specification.and(
                        buildSpecification(
                            criteria.getReviewId(),
                            root -> root.join(ReviewHistoryEntry_.review, JoinType.LEFT).get(Review_.id)
                        )
                    );
            }
            if (criteria.getReviewRatingId() != null) {
                specification =
                    specification.and(
                        buildSpecification(
                            criteria.getReviewRatingId(),
                            root -> root.join(ReviewHistoryEntry_.reviewRating, JoinType.LEFT).get(ReviewRating_.id)
                        )
                    );
            }
            if (criteria.getUserId() != null) {
                specification =
                    specification.and(
                        buildSpecification(criteria.getUserId(), root -> root.join(ReviewHistoryEntry_.user, JoinType.LEFT).get(User_.id))
                    );
            }
        }
        return specification;
    }

    protected Sort createSort(ReviewHistoryEntryCriteria criteria) {
        if (criteria == null || criteria.getSort() == null) {
            return Sort.unsorted();
        }
        final List<Sort.Order> sorting = new ArrayList<>();
        Arrays
            .stream(criteria.getSort())
            .forEach(spec -> {
                switch (spec) {
                    case "entryDate:desc":
                        sorting.add(Sort.Order.desc(ReviewHistoryEntry_.ENTRY_DATE));
                        break;
                    case "entryDate:asc":
                        sorting.add(Sort.Order.asc(ReviewHistoryEntry_.ENTRY_DATE));
                        break;
                    default:
                        log.warn("No sorting support for {}", spec);
                        break;
                }
            });
        return Sort.by(sorting);
    }
}
