import { TestBed } from '@angular/core/testing';
import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';

import { ISavedSearches } from '../saved-searches.model';
import { sampleWithRequiredData, sampleWithNewData, sampleWithPartialData, sampleWithFullData } from '../saved-searches.test-samples';

import { SavedSearchesService } from './saved-searches.service';

const requireRestSample: ISavedSearches = {
  ...sampleWithRequiredData,
};

/* eslint-disable @typescript-eslint/no-unsafe-return */
describe('SavedSearches Service', () => {
  let service: SavedSearchesService;
  let httpMock: HttpTestingController;
  let expectedResult: ISavedSearches | ISavedSearches[] | boolean | null;

  beforeEach(() => {
    TestBed.configureTestingModule({
      imports: [HttpClientTestingModule],
    });
    expectedResult = null;
    service = TestBed.inject(SavedSearchesService);
    httpMock = TestBed.inject(HttpTestingController);
  });

  describe('Service methods', () => {
    it('should find an element', () => {
      const returnedFromService = { ...requireRestSample };
      const expected = { ...sampleWithRequiredData };

      service.find(123).subscribe(resp => (expectedResult = resp.body));

      const req = httpMock.expectOne({ method: 'GET' });
      req.flush(returnedFromService);
      expect(expectedResult).toMatchObject(expected);
    });

    it('should create a SavedSearches', () => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      const savedSearches = { ...sampleWithNewData };
      const returnedFromService = { ...requireRestSample };
      const expected = { ...sampleWithRequiredData };

      service.create(savedSearches).subscribe(resp => (expectedResult = resp.body));

      const req = httpMock.expectOne({ method: 'POST' });
      req.flush(returnedFromService);
      expect(expectedResult).toMatchObject(expected);
    });

    it('should update a SavedSearches', () => {
      const savedSearches = { ...sampleWithRequiredData };
      const returnedFromService = { ...requireRestSample };
      const expected = { ...sampleWithRequiredData };

      service.update(savedSearches).subscribe(resp => (expectedResult = resp.body));

      const req = httpMock.expectOne({ method: 'PUT' });
      req.flush(returnedFromService);
      expect(expectedResult).toMatchObject(expected);
    });

    it('should partial update a SavedSearches', () => {
      const patchObject = { ...sampleWithPartialData };
      const returnedFromService = { ...requireRestSample };
      const expected = { ...sampleWithRequiredData };

      service.partialUpdate(patchObject).subscribe(resp => (expectedResult = resp.body));

      const req = httpMock.expectOne({ method: 'PATCH' });
      req.flush(returnedFromService);
      expect(expectedResult).toMatchObject(expected);
    });

    it('should return a list of SavedSearches', () => {
      const returnedFromService = { ...requireRestSample };

      const expected = { ...sampleWithRequiredData };

      service.query().subscribe(resp => (expectedResult = resp.body));

      const req = httpMock.expectOne({ method: 'GET' });
      req.flush([returnedFromService]);
      httpMock.verify();
      expect(expectedResult).toMatchObject([expected]);
    });

    it('should delete a SavedSearches', () => {
      const expected = true;

      service.delete(123).subscribe(resp => (expectedResult = resp.ok));

      const req = httpMock.expectOne({ method: 'DELETE' });
      req.flush({ status: 200 });
      expect(expectedResult).toBe(expected);
    });

    describe('addSavedSearchesToCollectionIfMissing', () => {
      it('should add a SavedSearches to an empty array', () => {
        const savedSearches: ISavedSearches = sampleWithRequiredData;
        expectedResult = service.addSavedSearchesToCollectionIfMissing([], savedSearches);
        expect(expectedResult).toHaveLength(1);
        expect(expectedResult).toContain(savedSearches);
      });

      it('should not add a SavedSearches to an array that contains it', () => {
        const savedSearches: ISavedSearches = sampleWithRequiredData;
        const savedSearchesCollection: ISavedSearches[] = [
          {
            ...savedSearches,
          },
          sampleWithPartialData,
        ];
        expectedResult = service.addSavedSearchesToCollectionIfMissing(savedSearchesCollection, savedSearches);
        expect(expectedResult).toHaveLength(2);
      });

      it("should add a SavedSearches to an array that doesn't contain it", () => {
        const savedSearches: ISavedSearches = sampleWithRequiredData;
        const savedSearchesCollection: ISavedSearches[] = [sampleWithPartialData];
        expectedResult = service.addSavedSearchesToCollectionIfMissing(savedSearchesCollection, savedSearches);
        expect(expectedResult).toHaveLength(2);
        expect(expectedResult).toContain(savedSearches);
      });

      it('should add only unique SavedSearches to an array', () => {
        const savedSearchesArray: ISavedSearches[] = [sampleWithRequiredData, sampleWithPartialData, sampleWithFullData];
        const savedSearchesCollection: ISavedSearches[] = [sampleWithRequiredData];
        expectedResult = service.addSavedSearchesToCollectionIfMissing(savedSearchesCollection, ...savedSearchesArray);
        expect(expectedResult).toHaveLength(3);
      });

      it('should accept varargs', () => {
        const savedSearches: ISavedSearches = sampleWithRequiredData;
        const savedSearches2: ISavedSearches = sampleWithPartialData;
        expectedResult = service.addSavedSearchesToCollectionIfMissing([], savedSearches, savedSearches2);
        expect(expectedResult).toHaveLength(2);
        expect(expectedResult).toContain(savedSearches);
        expect(expectedResult).toContain(savedSearches2);
      });

      it('should accept null and undefined values', () => {
        const savedSearches: ISavedSearches = sampleWithRequiredData;
        expectedResult = service.addSavedSearchesToCollectionIfMissing([], null, savedSearches, undefined);
        expect(expectedResult).toHaveLength(1);
        expect(expectedResult).toContain(savedSearches);
      });

      it('should return initial array if no SavedSearches is added', () => {
        const savedSearchesCollection: ISavedSearches[] = [sampleWithRequiredData];
        expectedResult = service.addSavedSearchesToCollectionIfMissing(savedSearchesCollection, undefined, null);
        expect(expectedResult).toEqual(savedSearchesCollection);
      });
    });

    describe('compareSavedSearches', () => {
      it('Should return true if both entities are null', () => {
        const entity1 = null;
        const entity2 = null;

        const compareResult = service.compareSavedSearches(entity1, entity2);

        expect(compareResult).toEqual(true);
      });

      it('Should return false if one entity is null', () => {
        const entity1 = { id: 123 };
        const entity2 = null;

        const compareResult1 = service.compareSavedSearches(entity1, entity2);
        const compareResult2 = service.compareSavedSearches(entity2, entity1);

        expect(compareResult1).toEqual(false);
        expect(compareResult2).toEqual(false);
      });

      it('Should return false if primaryKey differs', () => {
        const entity1 = { id: 123 };
        const entity2 = { id: 456 };

        const compareResult1 = service.compareSavedSearches(entity1, entity2);
        const compareResult2 = service.compareSavedSearches(entity2, entity1);

        expect(compareResult1).toEqual(false);
        expect(compareResult2).toEqual(false);
      });

      it('Should return false if primaryKey matches', () => {
        const entity1 = { id: 123 };
        const entity2 = { id: 123 };

        const compareResult1 = service.compareSavedSearches(entity1, entity2);
        const compareResult2 = service.compareSavedSearches(entity2, entity1);

        expect(compareResult1).toEqual(true);
        expect(compareResult2).toEqual(true);
      });
    });
  });

  afterEach(() => {
    httpMock.verify();
  });
});
